import unittest
from datetime import datetime
from unittest import mock

from pydantic import parse_raw_as

from tests.utils import mocked_requests, read_file
from traveltimepy.dto import Coordinates
from traveltimepy.dto.requests.zones import DepartureSearch, ArrivalSearch, Property
from traveltimepy.dto.responses.zones import DistrictsResponse, SectorsResponse
from traveltimepy.sdk import TravelTimeSdk
from traveltimepy.transportation import PublicTransport


class ZonesTest(unittest.TestCase):
    @mock.patch('requests.post', side_effect=mocked_requests)
    def test_districts(self, mock_post):
        sdk = TravelTimeSdk('appId', 'apiKey')

        departure_search = DepartureSearch(
            id='public transport from Trafalgar Square',
            departure_time=datetime(2022, 11, 24, 12, 0, 0),
            travel_time=200,
            coords=Coordinates(lat=51.507609, lng=-0.128315),
            reachable_postcodes_threshold=0.1,
            transportation=PublicTransport(),
            properties=[Property.TRAVEL_TIME_ALL, Property.TRAVEL_TIME_REACHABLE]
        )

        arrival_search = ArrivalSearch(
            id='public transport to Trafalgar Square',
            arrival_time=datetime(2022, 11, 24, 12, 0, 0),
            travel_time=200,
            coords=Coordinates(lat=51.507609, lng=-0.128315),
            reachable_postcodes_threshold=0.1,
            transportation=PublicTransport(),
            properties=[Property.COVERAGE]
        )

        response = sdk.districts([departure_search], [arrival_search])
        expected_response = parse_raw_as(DistrictsResponse, read_file('tests/resources/responses/districts.json'))
        self.assertEqual(response, expected_response)

    @mock.patch('requests.post', side_effect=mocked_requests)
    def test_sectors(self, mock_post):
        sdk = TravelTimeSdk('appId', 'apiKey')

        departure_search = DepartureSearch(
            id='public transport from Trafalgar Square',
            departure_time=datetime(2022, 11, 24, 12, 0, 0),
            travel_time=200,
            coords=Coordinates(lat=51.507609, lng=-0.128315),
            reachable_postcodes_threshold=0.1,
            transportation=PublicTransport(),
            properties=[Property.TRAVEL_TIME_ALL, Property.TRAVEL_TIME_REACHABLE]
        )

        arrival_search = ArrivalSearch(
            id='public transport to Trafalgar Square',
            arrival_time=datetime(2022, 11, 24, 12, 0, 0),
            travel_time=200,
            coords=Coordinates(lat=51.507609, lng=-0.128315),
            reachable_postcodes_threshold=0.1,
            transportation=PublicTransport(),
            properties=[Property.COVERAGE]
        )

        response = sdk.sectors([departure_search], [arrival_search])
        expected_response = parse_raw_as(SectorsResponse, read_file('tests/resources/responses/sectors.json'))
        self.assertEqual(response, expected_response)



from typing import List, Optional

from pydantic.main import BaseModel

from traveltimepy.dto import SearchId, LocationId
from traveltimepy.dto.responses import Fares


class Property(BaseModel):
    travel_time: int
    fares: Optional[Fares]


class Location(BaseModel):
    id: LocationId
    properties: List[Property]


class Result(BaseModel):
    search_id: SearchId
    locations: List[Location]
    unreachable: List[LocationId]


class RoutesResponse(BaseModel):
    results: List[Result]



from typing import Optional


class MockResponse:
    def __init__(self, data: Optional[str], status_code: int) -> None:
        self.data = data
        self.status_code = status_code

    @property
    def text(self) -> Optional[str]:
        return self.data


class UrlInfo:
    def __init__(self, request_file: Optional[str], response_file: str, status_code: int) -> None:
        self.request_file = request_file
        self.response_file = response_file
        self.status_code = status_code

    def check_request_data(self, request_data: str) -> bool:
        if self.request_file is not None:
            expected_request = read_file(f'tests/resources/requests/{self.request_file}').replace(' ', '')
            cur_request = request_data.replace(' ', '')
            return expected_request == cur_request
        else:
            return True

    def response_file_path(self) -> str:
        return f'tests/resources/responses/{self.response_file}'


urls = {
    'https://api.traveltimeapp.com/v4/map-info': UrlInfo(None, 'map_info.json', 200),
    'https://api.traveltimeapp.com/v4/time-map': UrlInfo('time_map.json', 'time_map.json', 200),
    'https://api.traveltimeapp.com/v4/routes': UrlInfo('routes.json', 'routes.json', 200),
    'https://api.traveltimeapp.com/v4/time-filter': UrlInfo('time_filter.json', 'time_filter.json', 200),
    'https://api.traveltimeapp.com/v4/geocoding/search': UrlInfo(None, 'geocoding.json', 200),
    'https://api.traveltimeapp.com/v4/geocoding/reverse': UrlInfo(None, 'geocoding.json', 200),
    'https://api.traveltimeapp.com/v4/time-filter/postcodes': UrlInfo('postcodes.json', 'postcodes.json', 200),
    'https://api.traveltimeapp.com/v4/time-filter/postcode-districts': UrlInfo('districts.json', 'districts.json', 200),
    'https://api.traveltimeapp.com/v4/time-filter/postcode-sectors': UrlInfo('sectors.json', 'sectors.json', 200),
    'https://api.traveltimeapp.com/v4/time-filter/fast': UrlInfo('time_filter_fast.json', 'time_filter_fast.json', 200),
    'https://api.traveltimeapp.com/v4/supported-locations': UrlInfo(
        'supported_locations.json',
        'supported_locations.json',
        200
    )
}


def mocked_requests(*args, **kwargs):
    url = kwargs.get('url')
    request_data = kwargs.get('data')
    url_info = urls.get(url)

    if url_info.check_request_data(request_data):
        response = read_file(url_info.response_file_path())
        return MockResponse(response, 200)
    else:
        return MockResponse(None, 404)


def check_request_data(path: str, data: str) -> bool:
    request = read_file(path)
    return data == request


def read_file(path: str) -> str:
    with open(path, 'r') as file:
        return file.read().replace('\n', '')



from datetime import datetime
from typing import List, Optional, Union

from pydantic.main import BaseModel

from traveltimepy.dto import SearchId, Location, LocationId
from traveltimepy.dto.requests import Property, FullRange
from traveltimepy.transportation import PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain


class ArrivalSearch(BaseModel):
    id: SearchId
    departure_location_ids: List[LocationId]
    arrival_location_id: LocationId
    arrival_time: datetime
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain]
    properties: List[Property]
    full_range: Optional[FullRange] = None


class DepartureSearch(BaseModel):
    id: SearchId
    arrival_location_ids: List[LocationId]
    departure_location_id: LocationId
    departure_time: datetime
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain]
    properties: List[Property]
    full_range: Optional[FullRange] = None


class RoutesRequest(BaseModel):
    locations: List[Location]
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]



import typing
from datetime import datetime

from typing import List, Optional

from pydantic.main import BaseModel

from traveltimepy.dto import Coordinates, SearchId
from traveltimepy.dto.requests import Range
from traveltimepy.transportation import PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain


class DepartureSearch(BaseModel):
    id: SearchId
    coords: Coordinates
    departure_time: datetime
    travel_time: int
    transportation: typing.Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain]
    range: Optional[Range] = None


class ArrivalSearch(BaseModel):
    id: SearchId
    coords: Coordinates
    arrival_time: datetime
    travel_time: int
    transportation: typing.Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain]
    range: Optional[Range] = None


class Intersection(BaseModel):
    id: SearchId
    search_ids: List[SearchId]


class Union(BaseModel):
    id: SearchId
    search_ids: List[SearchId]


class TimeMapRequest(BaseModel):
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]
    unions: List[Union]
    intersections: List[Intersection]



import math
from dataclasses import dataclass
from enum import Enum
from typing import List

from pydantic import BaseModel

from traveltimepy import TimeFilterFastRequest_pb2
from traveltimepy.dto import Coordinates


@dataclass
class TransportationInfo:
    code: int
    name: str


class Transportation(Enum):
    PUBLIC_TRANSPORT = TransportationInfo(0, 'pt')
    DRIVING = TransportationInfo(1, 'driving')
    DRIVING_FERRY = TransportationInfo(3, 'driving+ferry')
    WALKING = TransportationInfo(4, 'walking')
    CYCLING = TransportationInfo(5, 'cycling')
    CYCLING_FERRY = TransportationInfo(6, 'cycling+ferry')
    WALKING_FERRY = TransportationInfo(7, 'walking+ferry')


class Country(str, Enum):
    NETHERLANDS = 'nl'
    AUSTRIA = 'at'
    UNITED_KINGDOM = 'uk'
    BELGIUM = 'be'
    GERMANY = 'de'
    FRANCE = 'fr'
    IRELAND = 'ie'
    LITHUANIA = 'lt'


class OneToMany(BaseModel):
    origin_coordinates: Coordinates
    destination_coordinates: List[Coordinates]
    transportation: Transportation
    travel_time: int
    country: Country


class TimeFilterProtoRequest(BaseModel):
    one_to_many: OneToMany

    def to_proto(self) -> TimeFilterFastRequest_pb2.TimeFilterFastRequest:
        request = TimeFilterFastRequest_pb2.TimeFilterFastRequest()

        request.oneToManyRequest.departureLocation.lat = self.one_to_many.origin_coordinates.lat
        request.oneToManyRequest.departureLocation.lng = self.one_to_many.origin_coordinates.lng

        request.oneToManyRequest.transportation.type = self.one_to_many.transportation.value.code
        request.oneToManyRequest.travelTime = self.one_to_many.travel_time
        request.oneToManyRequest.arrivalTimePeriod = TimeFilterFastRequest_pb2.TimePeriod.WEEKDAY_MORNING

        mult = math.pow(10, 5)
        for destination in self.one_to_many.destination_coordinates:
            lat_delta = round((destination.lat - self.one_to_many.origin_coordinates.lat) * mult)
            lng_delta = round((destination.lng - self.one_to_many.origin_coordinates.lng) * mult)
            request.oneToManyRequest.locationDeltas.extend([lat_delta, lng_delta])

        return request



from enum import Enum

from pydantic import BaseModel


class Rectangle(BaseModel):
    min_lat: float
    max_lat: float
    min_lng: float
    max_lng: float


class Property(str, Enum):
    TRAVEL_TIME = 'travel_time'
    DISTANCE = 'distance'
    ROUTE = 'route'
    FARES = 'fares'


class FullRange(BaseModel):
    enabled: bool
    max_results: int
    width: int


class Range(BaseModel):
    enabled: bool
    width: int



import unittest
from unittest import mock

from tests.utils import mocked_requests, read_file
from datetime import datetime
from pydantic import parse_raw_as

from traveltimepy.dto import Coordinates
from traveltimepy.dto.requests import Range
from traveltimepy.dto.requests.time_map import DepartureSearch, ArrivalSearch, Union, Intersection
from traveltimepy.dto.responses.time_map import TimeMapResponse
from traveltimepy.sdk import TravelTimeSdk
from traveltimepy.transportation import PublicTransport, Driving


class TimeMapTest(unittest.TestCase):

    @mock.patch('requests.post', side_effect=mocked_requests)
    def test_time_map(self, mock_post):
        sdk = TravelTimeSdk('appId', 'apiKey')
        departure_search1 = DepartureSearch(
            id='search_1',
            coords=Coordinates(lat=51.507609, lng=-0.128315),
            departure_time=datetime(2022, 11, 24, 12, 0, 0),
            travel_time=900,
            transportation=PublicTransport()
        )
        departure_search2 = DepartureSearch(
            id='search_2',
            coords=Coordinates(lat=51.507609, lng=-0.128315),
            departure_time=datetime(2022, 11, 24, 12, 0, 0),
            travel_time=900,
            transportation=Driving()
        )
        arrival_search = ArrivalSearch(
            id='search_3',
            coords=Coordinates(lat=51.507609, lng=-0.128315),
            arrival_time=datetime(2022, 11, 24, 12, 0, 0),
            travel_time=900,
            transportation=PublicTransport(),
            range=Range(enabled=True, width=3600)
        )
        union = Union(
            id='search_4',
            search_ids=['search_2', 'search_3']
        )
        intersection = Intersection(
            id='search_5',
            search_ids=['search_2', 'search_3']
        )
        response = sdk.time_map(
            [arrival_search],
            [departure_search1, departure_search2],
            [union],
            [intersection]
        )
        expected_response = parse_raw_as(TimeMapResponse, read_file('tests/resources/responses/time_map.json'))
        self.assertEqual(response, expected_response)



from datetime import datetime
from enum import Enum
from typing import List, Union, Optional

from pydantic import BaseModel

from traveltimepy.dto import SearchId, Coordinates
from traveltimepy.dto.requests import FullRange
from traveltimepy.transportation import PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain


class Property(str, Enum):
    TRAVEL_TIME_REACHABLE = 'travel_time_reachable'
    TRAVEL_TIME_ALL = 'travel_time_all'
    COVERAGE = 'coverage'


class ArrivalSearch(BaseModel):
    id: SearchId
    coords: Coordinates
    travel_time: int
    arrival_time: datetime
    reachable_postcodes_threshold: float
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain]
    properties: List[Property]
    full_range: Optional[FullRange] = None


class DepartureSearch(BaseModel):
    id: SearchId
    coords: Coordinates
    travel_time: int
    departure_time: datetime
    reachable_postcodes_threshold: float
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain]
    properties: List[Property]
    full_range: Optional[FullRange] = None


class SectorsRequest(BaseModel):
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]


class DistrictsRequest(BaseModel):
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]


from datetime import datetime, time
from pydantic import BaseModel
from typing import Literal, NewType, List, Union, Optional

from traveltimepy.dto import Coordinates

PartId = NewType('PartId', str)


class BasicPart(BaseModel):
    id: PartId
    mode: str
    directions: str
    distance: int
    travel_time: int
    coords: List[Coordinates]
    type: Literal['basic']


class RoadPart(BaseModel):
    id: PartId
    mode: str
    directions: str
    distance: int
    travel_time: int
    coords: List[Coordinates]
    type: Literal['road']
    road: Optional[str]
    turn: Optional[str]


class StartEndPart(BaseModel):
    id: PartId
    mode: str
    directions: str
    distance: int
    travel_time: int
    coords: List[Coordinates]
    type: Literal['start_end']
    direction: str


class PublicTransportPart(BaseModel):
    id: PartId
    mode: str
    directions: str
    distance: int
    travel_time: int
    coords: List[Coordinates]
    line: str
    departure_station: str
    arrival_station: str
    departs_at: time
    arrives_at: time
    num_stops: int
    type: Literal['public_transport']


class Route(BaseModel):
    departure_time: datetime
    arrival_time: datetime
    parts: List[Union[BasicPart, PublicTransportPart, StartEndPart, RoadPart]]


class Ticket(BaseModel):
    type: str
    price: float
    currency: str


class FareBreakdown(BaseModel):
    modes: List[str]
    route_part_ids: List[int]
    tickets: List[Ticket]


class Fares(BaseModel):
    breakdown: List[FareBreakdown]
    tickets_total: List[Ticket]


import os
import unittest

from traveltimepy.dto import Coordinates

from traveltimepy.dto.requests.time_filter_proto import OneToMany, Transportation, Country
from traveltimepy.sdk import TravelTimeSdk


class TimeFilterProtoTest(unittest.TestCase):

    def test_proto(self):
        sdk = TravelTimeSdk(os.environ['PROTO_APP_ID'], os.environ['PROTO_API_KEY'])
        one_to_many = OneToMany(
            origin_coordinates=Coordinates(lat=51.425709, lng=-0.122061),
            destination_coordinates=[
                Coordinates(lat=51.348605, lng=-0.314783),
                Coordinates(lat=51.337205, lng=-0.315793)
            ],
            transportation=Transportation.DRIVING,
            travel_time=7200,
            country=Country.UNITED_KINGDOM
        )

        response = sdk.time_filter_proto(one_to_many)
        self.assertEqual(len(response.travel_times), 2)



import unittest

from unittest import mock

from pydantic.tools import parse_raw_as

from tests.utils import mocked_requests, read_file
from traveltimepy.dto import Location, Coordinates
from traveltimepy.dto.requests import Property
from traveltimepy.dto.requests.time_filter_fast import Transportation, ManyToOne, OneToMany
from traveltimepy.dto.responses.time_filter_fast import TimeFilterFastResponse
from traveltimepy.sdk import TravelTimeSdk


class TimeFilterFastTest(unittest.TestCase):

    @mock.patch('requests.post', side_effect=mocked_requests)
    def test_time_filter_fast(self, mock_post):
        sdk = TravelTimeSdk('appId', 'apiKey')
        locations = [
            Location(id='London center', coords=Coordinates(lat=51.508930, lng=-0.131387)),
            Location(id='Hyde Park', coords=Coordinates(lat=51.508824, lng=-0.167093)),
            Location(id='ZSL London Zoo', coords=Coordinates(lat=51.536067, lng=-0.153596))
        ]
        many_to_one = ManyToOne(
            id='many-to-one search example',
            departure_location_ids=['Hyde Park', 'ZSL London Zoo'],
            arrival_location_id='London center',
            transportation=Transportation(type='public_transport'),
            arrival_time_period='weekday_morning',
            travel_time=1900,
            properties=[Property.TRAVEL_TIME, Property.FARES]
        )
        one_to_many = OneToMany(
            id='one-to-many search example',
            arrival_location_ids=['Hyde Park', 'ZSL London Zoo'],
            departure_location_id='London center',
            transportation=Transportation(type='public_transport'),
            arrival_time_period='weekday_morning',
            travel_time=1900,
            properties=[Property.TRAVEL_TIME, Property.FARES]
        )

        response = sdk.time_filter_fast(locations, [many_to_one], [one_to_many])
        time_filter_response = read_file('tests/resources/responses/time_filter_fast.json')
        expected_response = parse_raw_as(TimeFilterFastResponse, time_filter_response)
        self.assertEqual(response, expected_response)



from typing import TypeVar, Type, Dict

import aiohttp
import requests
from pydantic.main import BaseModel
from pydantic.tools import parse_raw_as
from traveltimepy import AcceptType, TimeFilterFastResponse_pb2

from traveltimepy.dto.requests.time_filter_proto import TimeFilterProtoRequest
from traveltimepy.dto.responses.error import ResponseError
from traveltimepy.dto.responses.time_filter_proto import TimeFilterProtoResponse
from traveltimepy.errors import ApiError


T = TypeVar('T')
R = TypeVar('R')


async def send_post_request_async(response_class: Type[T], path: str, headers: Dict[str, str], body: BaseModel) -> T:
    url = '/'.join(['https://api.traveltimeapp.com', 'v4', path])
    async with aiohttp.ClientSession() as session:
        async with session.post(url=url, headers=headers, data=body.json()) as resp:
            body_text = await resp.text()
            return __process_response(response_class, resp.status, body_text)


async def send_get_request_async(
    response_class: Type[T],
    path: str,
    headers: Dict[str, str],
    params: Dict[str, str] = None
) -> T:
    url = '/'.join(['https://api.traveltimeapp.com', 'v4', path])
    async with aiohttp.ClientSession() as session:
        async with session.get(url=url, headers=headers, params=params) as resp:
            body_text = await resp.text()
            return __process_response(response_class, resp.status, body_text)


def send_proto_request(
    proto_request: TimeFilterProtoRequest,
    app_id: str,
    api_key: str
) -> TimeFilterProtoResponse:
    country = proto_request.one_to_many.country.value
    transport = proto_request.one_to_many.transportation.value.name
    url = '/'.join(['https://proto.api.traveltimeapp.com', 'api', 'v2', country, 'time-filter', 'fast', transport])

    resp = requests.post(
        url,
        headers={'Content-Type': AcceptType.OCTET_STREAM.value},
        data=proto_request.to_proto().SerializeToString(),
        auth=(app_id, api_key)
    )

    response_body = TimeFilterFastResponse_pb2.TimeFilterFastResponse()

    if resp.status_code != 200:
        msg = 'Travel Time API proto request failed with error code: {}\n'.format(resp.status_code)
        raise ApiError(msg)

    response_body.ParseFromString(resp.content)
    return TimeFilterProtoResponse(travel_times=response_body.properties.travelTimes[:])


def send_get_request(response_class: Type[T], path: str, headers: Dict[str, str], params: Dict[str, str] = None) -> T:
    url = '/'.join(['https://api.traveltimeapp.com', 'v4', path])
    resp = requests.get(url=url, headers=headers, params=params)
    return __process_response(response_class, resp.status_code, resp.text)


def send_post_request(response_class: Type[T], path: str, headers: Dict[str, str], body: BaseModel) -> T:
    url = '/'.join(['https://api.traveltimeapp.com', 'v4', path])
    resp = requests.post(url=url, headers=headers, data=body.json())
    return __process_response(response_class, resp.status_code, resp.text)


def __process_response(response_class: Type[T], status_code: int, text: str) -> T:
    if status_code != 200:
        parsed = parse_raw_as(ResponseError, text)
        msg = 'Travel Time API request failed \n{}\nError code: {}\n<{}>\n'.format(
            parsed.description,
            parsed.error_code,
            parsed.documentation_link
        )

        raise ApiError(msg)

    return parse_raw_as(response_class, text)



from traveltimepy import AcceptType, dto
from traveltimepy.dto import Location
from traveltimepy.dto.requests import time_map, time_filter, time_filter_proto, routes, postcodes, zones, Rectangle
from traveltimepy.dto.requests.postcodes import PostcodesRequest
from traveltimepy.dto.requests.routes import RoutesRequest
from traveltimepy.dto.requests.supported_locations import SupportedLocationsRequest
from traveltimepy.dto.requests.time_filter import TimeFilterRequest
from traveltimepy.dto.requests.time_filter_fast import ArrivalSearches, TimeFilterFastRequest, ManyToOne, OneToMany

from traveltimepy.dto.requests.time_map import *
from traveltimepy.dto.requests.zones import DistrictsRequest, SectorsRequest
from traveltimepy.dto.responses.map_info import MapInfoResponse
from traveltimepy.dto.responses.postcodes import PostcodesResponse
from traveltimepy.dto.responses.routes import RoutesResponse
from traveltimepy.dto.responses.supported_locations import SupportedLocationsResponse
from traveltimepy.dto.responses.time_filter import TimeFilterResponse
from traveltimepy.dto.responses.time_filter_fast import TimeFilterFastResponse
from traveltimepy.dto.responses.time_map import TimeMapResponse
from traveltimepy.dto.responses.zones import DistrictsResponse, SectorsResponse
from traveltimepy.utils import *

from geojson_pydantic import FeatureCollection


class TravelTimeSdk:

    def __init__(self, app_id: str, api_key: str) -> None:
        self.__app_id = app_id
        self.__api_key = api_key

    def time_filter_proto(self, one_to_many: time_filter_proto.OneToMany) -> TimeFilterProtoResponse:
        return send_proto_request(
            TimeFilterProtoRequest(one_to_many=one_to_many),
            self.__app_id,
            self.__api_key
        )

    def map_info(self) -> MapInfoResponse:
        return send_get_request(MapInfoResponse, 'map-info', self.__headers(AcceptType.JSON))

    async def map_info_async(self) -> MapInfoResponse:
        return await send_get_request_async(MapInfoResponse, 'map-info', self.__headers(AcceptType.JSON))

    def time_map(
        self,
        arrival_searches: List[dto.requests.time_map.ArrivalSearch],
        departure_searches: List[dto.requests.time_map.DepartureSearch],
        unions: List[Union] = [],
        intersections: List[Intersection] = []
    ) -> TimeMapResponse:
        return send_post_request(
            TimeMapResponse,
            'time-map',
            self.__headers(AcceptType.JSON),
            TimeMapRequest(
                departure_searches=departure_searches,
                arrival_searches=arrival_searches,
                unions=unions,
                intersections=intersections
            )
        )

    async def time_map_async(
        self,
        arrival_searches: List[dto.requests.time_map.ArrivalSearch],
        departure_searches: List[dto.requests.time_map.DepartureSearch],
        unions: List[Union] = [],
        intersections: List[Intersection] = []
    ) -> TimeMapResponse:
        return await send_post_request_async(
            TimeMapResponse,
            'time-map',
            self.__headers(AcceptType.JSON),
            TimeMapRequest(
                departure_searches=departure_searches,
                arrival_searches=arrival_searches,
                unions=unions,
                intersections=intersections
            )
        )

    def time_filter(
        self,
        locations: List[Location],
        departure_searches: List[dto.requests.time_filter.DepartureSearch],
        arrival_searches: List[dto.requests.time_filter.ArrivalSearch]
    ) -> TimeFilterResponse:
        return send_post_request(
            TimeFilterResponse,
            'time-filter',
            self.__headers(AcceptType.JSON),
            TimeFilterRequest(
                locations=locations,
                departure_searches=departure_searches,
                arrival_searches=arrival_searches
            )
        )

    async def time_filter_async(
        self,
        locations: List[Location],
        departure_searches: List[dto.requests.time_filter.DepartureSearch],
        arrival_searches: List[dto.requests.time_filter.ArrivalSearch]
    ) -> TimeFilterResponse:
        return await send_post_request_async(
            TimeFilterResponse,
            'time-filter',
            self.__headers(AcceptType.JSON),
            TimeFilterRequest(
                locations=locations,
                departure_searches=departure_searches,
                arrival_searches=arrival_searches
            )
        )

    def time_filter_fast(
        self,
        locations: List[Location],
        many_to_one: List[ManyToOne],
        one_to_many: List[OneToMany]
    ) -> TimeFilterFastResponse:
        return send_post_request(
            TimeFilterFastResponse,
            'time-filter/fast',
            self.__headers(AcceptType.JSON),
            TimeFilterFastRequest(
                locations=locations,
                arrival_searches=ArrivalSearches(many_to_one=many_to_one, one_to_many=one_to_many)
            )
        )

    async def time_filter_fast_async(
        self,
        locations: List[Location],
        many_to_one: List[ManyToOne],
        one_to_many: List[OneToMany]
    ) -> TimeFilterFastResponse:
        return await send_post_request_async(
            TimeFilterFastResponse,
            'time-filter/fast',
            self.__headers(AcceptType.JSON),
            TimeFilterFastRequest(
                locations=locations,
                arrival_searches=ArrivalSearches(many_to_one=many_to_one, one_to_many=one_to_many)
            )
        )

    def postcodes(
        self,
        departure_searches: List[postcodes.DepartureSearch],
        arrival_searches: List[postcodes.ArrivalSearch]
    ) -> PostcodesResponse:
        return send_post_request(
            PostcodesResponse,
            'time-filter/postcodes',
            self.__headers(AcceptType.JSON),
            PostcodesRequest(
                departure_searches=departure_searches,
                arrival_searches=arrival_searches
            )
        )

    def districts(
        self,
        departure_searches: List[zones.DepartureSearch],
        arrival_searches: List[zones.ArrivalSearch]
    ) -> DistrictsResponse:
        return send_post_request(
            DistrictsResponse,
            'time-filter/postcode-districts',
            self.__headers(AcceptType.JSON),
            DistrictsRequest(
                departure_searches=departure_searches,
                arrival_searches=arrival_searches
            )
        )

    def sectors(
        self,
        departure_searches: List[zones.DepartureSearch],
        arrival_searches: List[zones.ArrivalSearch]
    ) -> SectorsResponse:
        return send_post_request(
            SectorsResponse,
            'time-filter/postcode-sectors',
            self.__headers(AcceptType.JSON),
            SectorsRequest(
                departure_searches=departure_searches,
                arrival_searches=arrival_searches
            )
        )

    def routes(
        self,
        locations: List[Location],
        departure_searches: List[dto.requests.routes.DepartureSearch],
        arrival_searches: List[dto.requests.routes.ArrivalSearch]
    ) -> RoutesResponse:
        return send_post_request(
            RoutesResponse,
            'routes',
            self.__headers(AcceptType.JSON),
            RoutesRequest(
                locations=locations,
                departure_searches=departure_searches,
                arrival_searches=arrival_searches
            )
        )

    async def routes_async(
        self,
        locations: List[Location],
        departure_searches: List[dto.requests.time_filter.DepartureSearch],
        arrival_searches: List[dto.requests.time_filter.ArrivalSearch]
    ) -> RoutesResponse:
        return await send_post_request_async(
            RoutesResponse,
            'routes',
            self.__headers(AcceptType.JSON),
            RoutesRequest(
                locations=locations,
                departure_searches=departure_searches,
                arrival_searches=arrival_searches
            )
        )

    def geocoding(
        self,
        query: str,
        limit: Optional[int] = None,
        within_countries: Optional[List[str]] = None,
        format_name: Optional[bool] = None,
        format_exclude_country: Optional[bool] = None,
        rectangle: Optional[Rectangle] = None
    ) -> FeatureCollection:
        full_query = {
            'query': query,
            'limit': limit,
            'within.country': self.__combine_countries(within_countries),
            'format.name': format_name,
            'format.exclude.country': format_exclude_country,
            'bounds': self.__bounds(rectangle)
        }
        params = {key: str(value) for (key, value) in full_query.items() if value is not None}
        return send_get_request(FeatureCollection, 'geocoding/search', self.__headers(AcceptType.JSON), params)

    def geocoding_reverse(
        self,
        lat: float,
        lng: float,
        within_countries: Optional[List[str]] = None
    ) -> FeatureCollection:
        full_query = {
            'lat': lat,
            'lng': lng,
            'within.country': self.__combine_countries(within_countries)
        }
        params = {key: str(value) for (key, value) in full_query.items() if value is not None}
        return send_get_request(FeatureCollection, 'geocoding/reverse', self.__headers(AcceptType.JSON), params)

    def supported_locations(self, locations: List[Location]) -> SupportedLocationsResponse:
        return send_post_request(
            SupportedLocationsResponse,
            'supported-locations',
            self.__headers(AcceptType.JSON),
            SupportedLocationsRequest(locations=locations)
        )

    @staticmethod
    def __bounds(rectangle: Optional[Rectangle]) -> Optional[str]:
        if rectangle is not None:
            return f'{rectangle.min_lat},{rectangle.min_lng},{rectangle.max_lat},{rectangle.max_lng}'
        else:
            return None

    @staticmethod
    def __combine_countries(within_countries: Optional[List[str]]) -> Optional[str]:
        return ','.join(within_countries) if within_countries is not None and len(within_countries) != 0 else None

    def __headers(self, accept_type: AcceptType) -> Dict[str, str]:
        return {
            'X-Application-Id': self.__app_id,
            'X-Api-Key': self.__api_key,
            'Content-Type': 'application/json',
            'Accept': accept_type.value
        }



from typing import List

from pydantic import BaseModel

from traveltimepy.dto import SearchId, LocationId


class Ticket(BaseModel):
    type: str
    price: float
    currency: str


class Fares(BaseModel):
    tickets_total: List[Ticket]


class Properties(BaseModel):
    travel_time: int
    fares: Fares


class Location(BaseModel):
    id: LocationId
    properties: Properties


class Result(BaseModel):
    search_id: SearchId
    locations: List[Location]
    unreachable: List[LocationId]


class TimeFilterFastResponse(BaseModel):
    results: List[Result]



import unittest
from unittest import mock

from pydantic.tools import parse_raw_as

from tests.utils import mocked_requests, read_file
from traveltimepy.dto import Location, Coordinates
from traveltimepy.dto.responses.supported_locations import SupportedLocationsResponse
from traveltimepy.sdk import TravelTimeSdk


class SupportedLocationsTest(unittest.TestCase):

    @mock.patch('requests.post', side_effect=mocked_requests)
    def test_supported_locations(self, mock_get):
        sdk = TravelTimeSdk('appId', 'apiKey')
        locations = [
            Location(id='Kaunas', coords=Coordinates(lat=54.900008, lng=23.957734)),
            Location(id='London', coords=Coordinates(lat=51.506756, lng=-0.12805)),
            Location(id='Bangkok', coords=Coordinates(lat=13.761866, lng=100.544818)),
            Location(id='Lisbon', coords=Coordinates(lat=38.721869, lng=-9.138549)),
        ]
        response = sdk.supported_locations(locations)
        supported_locations = read_file('tests/resources/responses/supported_locations.json')
        expected_response = parse_raw_as(SupportedLocationsResponse, supported_locations)
        self.assertEqual(response, expected_response)



import unittest
from unittest import mock

from pydantic.tools import parse_raw_as

from tests.utils import mocked_requests, read_file
from traveltimepy.dto.responses.map_info import MapInfoResponse
from traveltimepy.sdk import TravelTimeSdk


class MapInfoTest(unittest.TestCase):

    @mock.patch('requests.get', side_effect=mocked_requests)
    def test_map_info(self, mock_get):
        sdk = TravelTimeSdk('appId', 'apiKey')

        response = sdk.map_info()
        expected_response = parse_raw_as(MapInfoResponse, read_file('tests/resources/responses/map_info.json'))
        self.assertEqual(response, expected_response)



from typing import List

from pydantic.main import BaseModel

from traveltimepy.dto import SearchId, Coordinates


class Shape(BaseModel):
    shell: List[Coordinates]
    holes: List[List[Coordinates]]


class Result(BaseModel):
    search_id: SearchId
    shapes: List[Shape]


class TimeMapResponse(BaseModel):
    results: List[Result]



from typing import Literal, List

from pydantic import BaseModel

from traveltimepy.dto import LocationId, Location
from traveltimepy.dto.requests import Property


class Transportation(BaseModel):
    type: Literal[
        'public_transport',
        'driving',
        'cycling',
        'walking',
        'walking+ferry',
        'cycling+ferry',
        'driving+ferry',
        'driving+public_transport'
    ]


class OneToMany(BaseModel):
    id: str
    departure_location_id: LocationId
    arrival_location_ids: List[LocationId]
    transportation: Transportation
    travel_time: int
    arrival_time_period: str
    properties: List[Property]


class ManyToOne(BaseModel):
    id: str
    arrival_location_id: LocationId
    departure_location_ids: List[LocationId]
    transportation: Transportation
    travel_time: int
    arrival_time_period: str
    properties: List[Property]


class ArrivalSearches(BaseModel):
    many_to_one: List[ManyToOne]
    one_to_many: List[OneToMany]


class TimeFilterFastRequest(BaseModel):
    locations: List[Location]
    arrival_searches: ArrivalSearches



from typing import List

from pydantic import BaseModel


class Property(BaseModel):
    travel_time: int
    distance: int


class Postcode(BaseModel):
    code: str
    properties: List[Property]


class Result(BaseModel):
    search_id: str
    postcodes: List[Postcode]


class PostcodesResponse(BaseModel):
    results: List[Result]



import unittest
from datetime import datetime

from unittest import mock

from pydantic.tools import parse_raw_as

from tests.utils import mocked_requests, read_file
from traveltimepy.dto import Location, Coordinates
from traveltimepy.dto.requests import FullRange, Property
from traveltimepy.dto.requests.time_filter import DepartureSearch, ArrivalSearch
from traveltimepy.dto.responses.time_filter import TimeFilterResponse
from traveltimepy.sdk import TravelTimeSdk
from traveltimepy.transportation import PublicTransport


class TimeFilterTest(unittest.TestCase):

    @mock.patch('requests.post', side_effect=mocked_requests)
    def test_time_filter(self, mock_post):
        sdk = TravelTimeSdk('appId', 'apiKey')
        locations = [
            Location(id='London center', coords=Coordinates(lat=51.508930, lng=-0.131387)),
            Location(id='Hyde Park', coords=Coordinates(lat=51.508824, lng=-0.167093)),
            Location(id='ZSL London Zoo', coords=Coordinates(lat=51.536067, lng=-0.153596))
        ]

        departure_search = DepartureSearch(
            id='forward search example',
            arrival_location_ids=['Hyde Park', 'ZSL London Zoo'],
            departure_location_id='London center',
            departure_time=datetime(2022, 11, 24, 12, 0, 0),
            travel_time=3600,
            transportation=PublicTransport(type='bus'),
            properties=[Property.TRAVEL_TIME],
            full_range=FullRange(enabled=True, max_results=3, width=600)
        )

        arrival_search = ArrivalSearch(
            id='backward search example',
            departure_location_ids=['Hyde Park', 'ZSL London Zoo'],
            arrival_location_id='London center',
            arrival_time=datetime(2022, 11, 24, 12, 0, 0),
            travel_time=3800,
            transportation=PublicTransport(type='bus'),
            properties=[Property.TRAVEL_TIME, Property.FARES, Property.ROUTE],
        )

        response = sdk.time_filter(locations, [departure_search], [arrival_search])
        expected_response = parse_raw_as(TimeFilterResponse, read_file('tests/resources/responses/time_filter.json'))
        self.assertEqual(response, expected_response)



from typing import List

from pydantic import BaseModel

from traveltimepy.dto import LocationId


class Location(BaseModel):
    id: LocationId
    map_name: str
    additional_map_names: List[str]


class SupportedLocationsResponse(BaseModel):
    locations: List[Location]
    unsupported_locations: List[LocationId]



import unittest
from unittest import mock

from geojson_pydantic import FeatureCollection
from pydantic.tools import parse_raw_as

from tests.utils import mocked_requests, read_file
from traveltimepy.sdk import TravelTimeSdk


class GeocodingTest(unittest.TestCase):

    @mock.patch('requests.get', side_effect=mocked_requests)
    def test_geocoding_search(self, mock_get):
        sdk = TravelTimeSdk('appId', 'apiKey')

        response = sdk.geocoding(query='Parliament square', limit=30)
        expected_response = parse_raw_as(FeatureCollection, read_file('tests/resources/responses/geocoding.json'))
        self.assertEqual(response, expected_response)

    @mock.patch('requests.get', side_effect=mocked_requests)
    def test_geocoding_reverse(self, mock_get):
        sdk = TravelTimeSdk('appId', 'apiKey')

        response = sdk.geocoding_reverse(lat=51.507281, lng=-0.132120)
        expected_response = parse_raw_as(FeatureCollection, read_file('tests/resources/responses/geocoding.json'))
        self.assertEqual(response, expected_response)



from datetime import datetime
from typing import List, Union, Optional

from pydantic import BaseModel

from traveltimepy.dto import SearchId, Coordinates
from traveltimepy.dto.requests import Property, FullRange
from traveltimepy.transportation import PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain


class ArrivalSearch(BaseModel):
    id: SearchId
    coords: Coordinates
    travel_time: int
    arrival_time: datetime
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain]
    properties: List[Property]
    full_range: Optional[FullRange] = None


class DepartureSearch(BaseModel):
    id: SearchId
    coords: Coordinates
    travel_time: int
    departure_time: datetime
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain]
    properties: List[Property]
    full_range: Optional[FullRange] = None


class PostcodesRequest(BaseModel):
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]



from typing import NewType

from pydantic.main import BaseModel

SearchId = NewType('SearchId', str)
LocationId = NewType('LocationId', str)


class Coordinates(BaseModel):
    lat: float
    lng: float


class Location(BaseModel):
    id: LocationId
    coords: Coordinates
